from typing import Any, Callable, Iterable, Type

from werkzeug.datastructures import FileStorage


class UploadFile(FileStorage):
    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        yield cls.validate

    @classmethod
    def validate(cls: Type["UploadFile"], v: Any) -> Any:
        if not isinstance(v, FileStorage):
            raise ValueError(f"Expected UploadFile, received: {type(v)}")
        return v

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(format="binary", type="string")



from inspect import getdoc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from flask import current_app, render_template_string
from pydantic import BaseModel
from pydantic.fields import ModelField, Undefined
from pydantic.schema import (
    TypeModelOrEnum,
    TypeModelSet,
    default_ref_template,
    get_flat_models_from_models,
    get_model_name_map,
    model_process_schema,
    schema,
)

from flask_sugar.constans import ALLOW_METHODS, REF_PREFIX, REF_TEMPLATE
from flask_sugar.templates import redoc_template, swagger_template
from flask_sugar.view import ParameterInfo, View

if TYPE_CHECKING:
    from flask_sugar.app import Sugar

    current_app: Sugar


def get_openapi_json(
    openapi_version: str,
    title: str,
    version: str,
    tags: Optional[List[Dict[str, Any]]] = None,
    description: Optional[str] = "",
    terms_service: Optional[str] = None,
    contact: Optional[Dict[str, str]] = None,
    license_: Optional[Dict[str, str]] = None,
    servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
    paths: Optional[Dict[str, Union[str, Any]]] = None,
    components: Optional[Dict[str, Union[str, Any]]] = None,
) -> Dict[str, Any]:
    info: Dict[str, Any] = {"title": title, "version": version}
    if description:
        info["description"] = description
    if terms_service:
        info["termsOfService"] = terms_service
    if contact:
        info["contact"] = contact
    if license_:
        info["license"] = license_
    source = {
        "openapi": openapi_version,
        "info": info,
    }
    if tags:
        source["tags"] = tags
    if servers:
        source["servers"] = servers
    if paths:
        source["paths"] = paths
    if components:
        source["components"] = components
    return source


def openapi_json_view() -> Dict[str, Any]:
    paths, components = collect_paths_components()

    return get_openapi_json(
        openapi_version=current_app.openapi_version,
        title=current_app.title,
        version=current_app.doc_version,
        tags=current_app.tags,
        description=current_app.description,
        terms_service=current_app.terms_service,
        contact=current_app.contact,
        license_=current_app.license_,
        servers=current_app.servers,
        paths=paths,
        components=components,
    )


def swagger() -> str:
    return render_template_string(
        swagger_template,
        openapi_json_url=current_app.openapi_json_url,
        title=current_app.title + " Swagger",
        swagger_js_url=current_app.swagger_js_url,
        swagger_css_url=current_app.swagger_css_url,
    )


def redoc() -> str:
    return render_template_string(
        redoc_template,
        openapi_json_url=current_app.openapi_json_url,
        title=current_app.title + " Redoc",
        redoc_js_url=current_app.redoc_js_url,
    )


def get_parameters(
    model: Optional[Type[BaseModel]], parameter_infos: List[ParameterInfo]
) -> List[Dict[str, Any]]:
    if not model:
        return []
    model_fields: Dict[str, ModelField] = model.__fields__
    properties = model.schema(ref_template=REF_TEMPLATE)["properties"]
    parameters = []
    for parameter_info in parameter_infos:
        model_field = model_fields[parameter_info.name]
        field_info = parameter_info.parameter.field_info
        doc_parameter = {
            "name": model_field.alias,
            "in": parameter_info.parameter.in_,
            "schema": properties[model_field.alias],
            "required": model_field.required,
        }

        if field_info.description:
            doc_parameter["description"] = field_info.description
        if parameter_info.parameter.examples:
            doc_parameter["examples"] = parameter_info.parameter.examples
        elif parameter_info.parameter.example != Undefined:
            doc_parameter["example"] = parameter_info.parameter.example
        if parameter_info.parameter.deprecated:
            doc_parameter["deprecated"] = parameter_info.parameter.deprecated
        parameters.append(doc_parameter)
    return parameters


def get_flat_models_from_views(view_functions: Iterable[Callable]) -> TypeModelSet:
    models = []
    for view in view_functions:
        if not isinstance(view, View) and getattr(view, "doc_enable"):
            continue
        if view.ParamModel:
            models.append(view.ParamModel)
        if view.body_info or view.FormModel:
            if view.FormModel:
                models.append(view.FormModel)
            else:
                models.append(view.body_info.model)  # type:ignore
        if view.response_model:
            models.append(view.response_model)
    return get_flat_models_from_models(models)


def schema(
    models: TypeModelSet,
    *,
    model_name_map: Dict[TypeModelOrEnum, str],
    by_alias: bool = True,
    title: Optional[str] = None,
    description: Optional[str] = None,
    ref_prefix: Optional[str] = None,
    ref_template: str = default_ref_template,
) -> Dict[str, Any]:
    definitions = {}
    output_schema: Dict[str, Any] = {}
    if title:
        output_schema["title"] = title
    if description:
        output_schema["description"] = description
    for model in models:
        m_schema, m_definitions, m_nested_models = model_process_schema(
            model,
            by_alias=by_alias,
            model_name_map=model_name_map,
            ref_prefix=ref_prefix,
            ref_template=ref_template,
        )
        definitions.update(m_definitions)
        model_name = model_name_map[model]
        definitions[model_name] = m_schema
    if definitions:
        output_schema["definitions"] = definitions
    return output_schema


def collect_paths_components() -> Tuple[Dict[str, Any], Dict[str, Any]]:
    paths = {}
    components = {}
    flat_models = get_flat_models_from_views(current_app.view_functions.values())
    model_name_map = get_model_name_map(flat_models)
    schemas = schema(
        flat_models, model_name_map=model_name_map, ref_prefix=REF_PREFIX
    ).get("definitions", {})
    for rule in current_app.url_map.iter_rules():
        view: View = cast(View, current_app.view_functions[rule.endpoint])
        if not getattr(view, "doc_enable"):
            continue
        path_item = {}
        for method in rule.methods:
            operation = {}
            method: str = method.lower()
            if method not in ALLOW_METHODS:
                continue
            if current_app.doc_route_filter and not current_app.doc_route_filter(
                view, rule
            ):
                continue
            parameters = get_parameters(view.ParamModel, view.parameter_infos)
            if parameters:
                operation["parameters"] = parameters

            if view.tags:
                operation["tags"] = view.tags

            summary = view.summary or view.view_func.__name__.replace("_", " ").title()
            if summary:
                operation["summary"] = summary

            description = view.description or getdoc(view)
            if description:
                operation["description"] = description

            if view.deprecated:
                operation["deprecated"] = view.deprecated

            operation_id = view.operation_id or rule.endpoint + "__" + method
            if operation_id:
                operation["operationId"] = operation_id

            if view.security:
                operation["security"] = view.security

            path_item[method] = operation
            if view.body_info or view.FormModel:
                if view.FormModel:
                    body_model_name = model_name_map[view.FormModel]
                    media_type = "multipart/form-data"
                else:
                    body_model_name = model_name_map[
                        view.body_info.model  # type:ignore
                    ]
                    media_type = view.body_info.parameter.media_type  # type:ignore

                operation["requestBody"] = {
                    "content": {
                        media_type: {"schema": {"$ref": REF_PREFIX + body_model_name}}
                    },
                    "required": True,
                }
            response_schema = {}
            if view.response_model:
                response_schema["$ref"] = (
                    REF_PREFIX + model_name_map[view.response_model]
                )

            responses: Dict[Union[int, str], Dict[str, Any]] = {
                "200": {
                    "description": view.response_description,
                    "content": {"application/json": {"schema": response_schema}},
                }
            }

            if view.responses:
                responses.update(view.responses)
            operation["responses"] = responses
            if view.extra:
                operation.update(view.extra)
        paths.setdefault(view.path, {}).update(path_item)

        if schemas:
            components["schemas"] = schemas
    if current_app.security_schemes:
        components["securitySchemes"] = current_app.security_schemes

    return paths, components



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union

from flask import Blueprint as _Blueprint
from pydantic import BaseModel

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


class Blueprint(_Blueprint):
    def __init__(
        self,
        name: str,
        import_name: str,
        static_folder: Optional[str] = None,
        static_url_path: Optional[str] = None,
        template_folder: Optional[str] = None,
        url_prefix: Optional[str] = None,
        subdomain: Optional[str] = None,
        url_defaults: Optional[dict] = None,
        root_path: Optional[str] = None,
        cli_group: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
    ):
        super().__init__(
            name,
            import_name,
            static_folder=static_folder,
            static_url_path=static_url_path,
            template_folder=template_folder,
            url_prefix=url_prefix,
            subdomain=subdomain,
            url_defaults=url_defaults,
            root_path=root_path,
            cli_group=cli_group,
        )
        self.tags: List[str] = [name] if tags is None else tags
        self.deprecated = deprecated

    def add_url_rule(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        view_func: Optional[Callable] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> None:
        """Like :meth:`Flask.add_url_rule` but for a blueprint.  The endpoint for
        the :func:`url_for` function is prefixed with the name of the blueprint.
        """
        if endpoint and "." in endpoint:
            raise ValueError("'endpoint' may not contain a dot '.' character.")

        if view_func and hasattr(view_func, "__name__") and "." in view_func.__name__:
            raise ValueError("'view_func' name may not contain a dot '.' character.")

        if extends_tags:
            tags = self.tags.copy() + (tags or [])

        self.record(
            lambda s: s.add_url_rule(
                rule,
                endpoint,
                view_func,
                provide_automatic_options=provide_automatic_options,
                doc_enable=doc_enable,
                tags=tags,
                summary=summary,
                description=description,
                response_model=response_model,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated or self.deprecated,
                operation_id=operation_id,
                security=security,
                extra=extra,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                **options,
            )
        )

    def get(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().get(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            extends_tags=extends_tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def post(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().post(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            extends_tags=extends_tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def put(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().put(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            extends_tags=extends_tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def delete(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().delete(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            extends_tags=extends_tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def patch(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        extends_tags: bool = True,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().patch(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            extends_tags=extends_tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )



swagger_template: str = """
<!DOCTYPE html>
<html>

<head>
    <link type="text/css" rel="stylesheet" href="{{ swagger_css_url }}">
    <title>{{ title }}</title>
</head>

<body>
    <div id="swagger-ui">
    </div>
    <script src="{{ swagger_js_url }}"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
        const ui = SwaggerUIBundle({
            url: "{{ url_for('openapi.openapi_json_view') }}",
            dom_id: '#swagger-ui',
            presets: [
                SwaggerUIBundle.presets.apis,
                SwaggerUIBundle.SwaggerUIStandalonePreset
            ],
            layout: "BaseLayout",
            deepLinking: true,
            showExtensions: true,
            showCommonExtensions: true
        })
    </script>
</body>

</html>
"""
redoc_template: str = """
<!DOCTYPE html>
<html>

<head>
    <title>{{ title }}</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1">

    <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
    <!--
    ReDoc doesn't change outer page styles
    -->
    <style>
        body {
            margin: 0;
            padding: 0;
        }
    </style>
</head>

<body>
    <redoc spec-url="{{ url_for('openapi.openapi_json_view') }}"></redoc>
    <script src="{{ redoc_js_url }}"> </script>
</body>

</html>
"""



from functools import update_wrapper
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from flask import request
from flask.typing import ResponseReturnValue
from pydantic import (
    BaseModel,
    ValidationError,
    create_model,
    create_model_from_typeddict,
)
from pydantic.fields import FieldInfo, ModelField
from typing_extensions import Literal
from werkzeug.datastructures import ImmutableMultiDict

from flask_sugar import params
from flask_sugar.datastructures import UploadFile
from flask_sugar.exceptions import RequestValidationError
from flask_sugar.utils import (
    get_list_value,
    get_long_obj_name,
    get_param_annotation,
    get_path_param_names,
    get_typed_signature,
    is_list_type,
    is_subclass,
    is_typed_dict,
)

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


class BodyInfo(NamedTuple):
    name: str
    model: Type[BaseModel]
    parameter: params.Body


ParamType = TypeVar("ParamType", params.Param, params.File)


class ParameterInfo(Generic[ParamType]):
    def __init__(self, name: str, is_list: bool, parameter: ParamType) -> None:
        self.name = name
        self.is_list = is_list
        self.parameter = parameter


class View:
    """wrap view_func"""

    def __init__(
        self,
        path: str,
        view_func: Callable,
        endpoint: Optional[str] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "success",
        response_model: Optional[Type[BaseModel]] = None,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
    ) -> None:

        self.path = path
        self.view_func = view_func
        update_wrapper(self, view_func)  # type:ignore
        self.doc_enable = doc_enable
        self.tags = tags
        self.summary = summary
        self.description = description
        self.response_description = response_description
        self.response_model = response_model
        self.responses = responses
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.security = security
        self.extra = extra
        self.ParamModel: Optional[Type[BaseModel]] = None
        self.FormModel: Optional[Type[BaseModel]] = None
        self.FileModel: Optional[Type[BaseModel]] = None
        self.parameter_infos: List[ParameterInfo[params.Param]] = []
        self.file_infos: List[ParameterInfo[params.File]] = []
        self.body_info: Optional[BodyInfo] = None
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none

        field_definitions: Dict[str, Tuple[Any, FieldInfo]] = {}
        path_param_names = get_path_param_names(path)
        signature = get_typed_signature(view_func)
        file_definitions: Dict[str, Tuple[Any, FieldInfo]] = {}
        if not response_model:
            if is_typed_dict(signature.return_annotation):
                self.response_model = create_model_from_typeddict(
                    signature.return_annotation,
                    __module__=signature.return_annotation.__module__,
                )
            elif is_subclass(signature.return_annotation, BaseModel):
                self.response_model = signature.return_annotation

        for param_name, param in signature.parameters.items():
            if param.kind in (param.VAR_KEYWORD, param.VAR_POSITIONAL):
                continue
            annotation = get_param_annotation(param)
            if is_subclass(annotation, BaseModel):
                assert (
                    self.body_info is None
                ), "a view_func require only one BaseModel field"
                if param.default == param.empty:
                    parameter = params.Body(...)
                else:
                    assert isinstance(
                        param.default, params.Body
                    ), "body field default value must be instance of params.Body"
                    parameter = param.default

                self.body_info = BodyInfo(
                    name=param_name,
                    model=annotation,
                    parameter=parameter,
                )
                continue
            if isinstance(param.default, params.File):
                annotation = param.annotation
                if param.annotation == param.empty:
                    annotation = UploadFile
                is_list = is_list_type(annotation)
                self.file_infos.append(
                    ParameterInfo(
                        name=param.name,
                        is_list=is_list,
                        parameter=param.default,
                    )
                )
                file_definitions[param_name] = (annotation, param.default.field_info)
                continue

            if param_name in path_param_names:
                assert (param.default == param.empty) or isinstance(
                    param.default, params.Path
                ), "path param default value must be subclass of params.Path or empty"
                if param.default == param.empty:
                    parameter = params.Path(...)
                else:
                    parameter = param.default
            else:
                if param.default == param.empty:
                    parameter = params.Query(...)
                elif isinstance(param.default, params.Param):
                    parameter = param.default
                else:
                    parameter = params.Query(default=param.default)
            field_definitions[param_name] = (annotation, parameter.field_info)
            self.parameter_infos.append(
                ParameterInfo(
                    name=param_name,
                    is_list=is_list_type(annotation),
                    parameter=parameter,
                )
            )
        if field_definitions:
            self.ParamModel = create_model(
                get_long_obj_name(view_func, f"{endpoint or ''}__ParamModel"),
                **field_definitions,
            )

        if self.file_infos:
            if self.body_info:
                assert isinstance(
                    self.body_info.parameter, params.Form
                ), "file field cant not coexist with body field"
                base_model = self.body_info.model
            else:
                base_model = None
            self.FileModel = create_model(
                get_long_obj_name(view_func, "FileModel"), **file_definitions
            )
            self.FormModel = create_model(
                get_long_obj_name(view_func, "FormModel"),
                __base__=base_model,
                **file_definitions,
            )

    @staticmethod
    def get_value(
        in_: Literal["query", "header", "path", "cookie", "file"],
        alias: str,
        param_name: str,
        is_list: bool,
        kwargs: Dict[str, Any],
    ) -> Any:
        if in_ == "query":
            return request.args.getlist(alias) if is_list else request.args.get(alias)
        if in_ == "header":
            rtn = request.headers.get(alias)
            return get_list_value(rtn) if is_list else rtn
        if in_ == "cookie":
            rtn = request.cookies.get(alias)
            return get_list_value(rtn) if is_list else rtn
        if in_ == "file":
            return request.files.getlist(alias) if is_list else request.files.get(alias)
        return kwargs[param_name]

    def get_request_values(
        self,
        parameter_infos: List[ParameterInfo],
        ParamModel: Type[BaseModel],
        kwargs: Dict[str, Any],
        result_use_alias: bool = True,
    ) -> Dict[str, Any]:
        values = {}
        model_fields: Dict[str, ModelField] = ParamModel.__fields__
        for parameter in parameter_infos:
            model_field = model_fields[parameter.name]
            in_ = parameter.parameter.in_
            alias = model_field.alias
            value = self.get_value(
                in_, alias, parameter.name, parameter.is_list, kwargs
            )
            if value is not None:
                if result_use_alias:
                    values[alias] = value
                else:
                    values[parameter.name] = value
        return values

    def inject_data(
        self, kwargs: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], List[Dict[str, Any]]]:
        errors: List[Dict[str, Any]] = []
        if self.ParamModel:
            request_values = self.get_request_values(
                self.parameter_infos, self.ParamModel, kwargs
            )
            try:
                param_data = self.ParamModel(**request_values)
                kwargs.update(param_data.dict())
            except ValidationError as e:
                errors.extend(e.errors())

        if self.body_info:
            body_values = getattr(request, self.body_info.parameter.request_attr) or {}
            if isinstance(body_values, ImmutableMultiDict):
                body_values = body_values.to_dict()
            try:
                kwargs[self.body_info.name] = self.body_info.model(**body_values)
            except ValidationError as e:
                errors.extend(e.errors())

        if self.FileModel:
            files = self.get_request_values(
                self.file_infos, self.FileModel, kwargs, False
            )

            try:
                file_model = self.FileModel(**files)
                kwargs.update(file_model.dict())
            except ValidationError as e:
                errors.extend(e.errors())

        return kwargs, errors

    def create_response(
        self, response: Union[ResponseReturnValue, BaseModel]
    ) -> ResponseReturnValue:
        if isinstance(response, BaseModel):
            return response.dict()
        if isinstance(response, dict) and self.response_model:
            return self.response_model(**response).dict(
                include=self.response_model_include,
                exclude=self.response_model_exclude,
                by_alias=self.response_model_by_alias,
                exclude_unset=self.response_model_exclude_unset,
                exclude_defaults=self.response_model_exclude_defaults,
                exclude_none=self.response_model_exclude_none,
            )
        return response

    def __call__(self, **kwargs) -> Any:
        if self.view_func is None:
            return self.view_func
        cleaned_data, errors = self.inject_data(kwargs)
        if errors:
            raise RequestValidationError(errors)
        response = self.view_func(**cleaned_data)
        return self.create_response(response)

    def __repr__(self):
        return f"View(view_func={self.view_func}, doc_enable={self.doc_enable})"

    def __eq__(self, other: Any) -> bool:
        other_view_func = getattr(other, "view_func", None)
        return self.view_func == other_view_func



from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union

from flask import Flask
from pydantic import BaseModel
from werkzeug.routing import Rule

from flask_sugar.blueprints import Blueprint
from flask_sugar.errorhandlers import validation_error_handler
from flask_sugar.exceptions import RequestValidationError
from flask_sugar.openapi import openapi_json_view, redoc, swagger
from flask_sugar.utils import convert_path
from flask_sugar.view import View

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


class Sugar(Flask):
    def __init__(
        self,
        import_name: str,
        static_url_path: Optional[str] = None,
        static_folder: Optional[str] = "static",
        static_host: Optional[str] = None,
        host_matching: bool = False,
        subdomain_matching: bool = False,
        template_folder: Optional[str] = "templates",
        instance_path: Optional[str] = None,
        instance_relative_config: bool = False,
        root_path: Optional[str] = None,
        title: str = "FlaskSugar",
        doc_version: str = "0.1.0",
        description: Optional[str] = None,
        terms_service: Optional[str] = None,
        contact: Optional[Dict[str, str]] = None,
        license_: Optional[Dict[str, str]] = None,
        servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
        tags: Optional[List[Dict[str, Any]]] = None,
        security_schemes: Optional[Dict[str, Any]] = None,
        enable_doc: bool = True,
        cache_openapi_json: bool = True,
        openapi_url_prefix: Optional[str] = None,
        openapi_json_url: str = "/openapi.json",
        swagger_url: str = "/doc",
        redoc_url: str = "/redoc",
        swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js",
        swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css",
        redoc_js_url: str = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js",
        default_validation_errorhandler: Optional[Callable[..., Any]] = None,
        doc_route_filter: Optional[Callable[[View, Rule], bool]] = None,
    ):
        super().__init__(
            import_name=import_name,
            static_url_path=static_url_path,
            static_folder=static_folder,
            static_host=static_host,
            host_matching=host_matching,
            subdomain_matching=subdomain_matching,
            template_folder=template_folder,
            instance_path=instance_path,
            instance_relative_config=instance_relative_config,
            root_path=root_path,
        )
        self.title = title
        self.doc_version = doc_version
        self.description = description
        self.terms_service = terms_service
        self.contact = contact
        self.license_ = license_
        self.servers = servers
        self.tags = tags
        self.security_schemes = security_schemes
        self.cache_openapi_json = cache_openapi_json
        self.openapi_url_prefix = openapi_url_prefix
        self.openapi_json_url = openapi_json_url
        self.swagger_url = swagger_url
        self.redoc_url = redoc_url
        self.openapi_version = "3.0.2"
        self.swagger_js_url = swagger_js_url
        self.swagger_css_url = swagger_css_url
        self.redoc_js_url = redoc_js_url
        self.doc_route_filter = doc_route_filter
        error_handler = (
            default_validation_errorhandler
            if default_validation_errorhandler is not None
            else validation_error_handler
        )
        self.register_error_handler(RequestValidationError, error_handler)
        if enable_doc:
            self.init_doc()

    def add_url_rule(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        view_func: Optional[Callable] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> None:
        path = convert_path(rule)
        assert view_func, "view_func can't be None"
        if endpoint == "static":
            doc_enable = False
        view = View(
            path=path,
            view_func=view_func,
            endpoint=endpoint,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            response_model=response_model,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
        )
        super().add_url_rule(rule, endpoint, view, provide_automatic_options, **options)

    def init_doc(self):
        openapi_bp = Blueprint("openapi", __name__, url_prefix=self.openapi_url_prefix)
        if self.openapi_json_url:
            _openapi_json_view = (
                lru_cache(maxsize=1)(openapi_json_view)
                if self.cache_openapi_json
                else openapi_json_view
            )
            openapi_bp.add_url_rule(
                self.openapi_json_url, view_func=_openapi_json_view, doc_enable=False
            )

        if self.openapi_json_url and self.swagger_url:
            openapi_bp.add_url_rule(
                self.swagger_url, view_func=swagger, doc_enable=False
            )

        if self.openapi_json_url and self.redoc_url:
            openapi_bp.add_url_rule(self.redoc_url, view_func=redoc, doc_enable=False)

        self.register_blueprint(openapi_bp)

    def get(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().get(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def post(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().post(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def put(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().put(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def delete(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().delete(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )

    def patch(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        provide_automatic_options: Optional[bool] = None,
        doc_enable: bool = True,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_model: Optional[Type[BaseModel]] = None,
        response_description: str = "success",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        security: Optional[List[Dict[str, Any]]] = None,
        extra: Optional[Dict[str, Any]] = None,
        response_model_include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        **options: Any,
    ) -> Callable:
        return super().patch(
            rule=rule,
            endpoint=endpoint,
            provide_automatic_options=provide_automatic_options,
            doc_enable=doc_enable,
            tags=tags,
            summary=summary,
            description=description,
            response_model=response_model,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            security=security,
            extra=extra,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            **options,
        )



# coding = utf-8
import re

from setuptools import find_packages, setup

with open("flask_sugar/__init__.py", "r") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(  # type:ignore
        1
    )
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="flask-sugar",
    python_requires=">=3.6",
    version=version,
    description="Flask Sugar is a web framework for building APIs with Flask, Pydantic and Python 3.6+ type hints.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="shangsky",
    author_email="t_c_y@outlook.com",
    maintainer="shangsky",
    maintainer_email="t_c_y@outlook.com",
    license="MIT",
    packages=find_packages(),
    platforms=["all"],
    url="https://github.com/shangsky/flask-sugar",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=["flask>=2.0", "pydantic>=1.8,<2.0.0"],
)



from flask_sugar.app import Sugar
from flask_sugar.blueprints import Blueprint
from flask_sugar.datastructures import UploadFile
from flask_sugar.exceptions import RequestValidationError
from flask_sugar.param_functions import Body, Cookie, File, Form, Header, Path, Query

__version__ = "0.0.15"

__all__ = [
    "Sugar",
    "Blueprint",
    "UploadFile",
    "RequestValidationError",
    "Body",
    "Cookie",
    "File",
    "Form",
    "Header",
    "Path",
    "Query",
]
